/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/* librvngabw
 * Version: MPL 2.0 / LGPLv2.1+
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Major Contributor(s):
 * Copyright (C) 2002-2004 William Lachance (wrlach@gmail.com)
 * Copyright (C) 2004 Fridrich Strba (fridrich.strba@bluewin.ch)
 *
 * For minor contributions see the git repository.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU Lesser General Public License Version 2.1 or later
 * (LGPLv2.1+), in which case the provisions of the LGPLv2.1+ are
 * applicable instead of those above.
 *
 * For further information visit http://libwpd.sourceforge.net
 */

/* "This product is not manufactured, approved, or supported by
 * Corel Corporation or Corel Corporation Limited."
 */

#include <math.h>
#include <time.h>

#include <locale>
#include <string>
#include <sstream>

#include "FilterInternal.hxx"
#include "DocumentElement.hxx"
#include "TextRunStyle.hxx"

#include "Table.hxx"
#ifdef _MSC_VER
#include <minmax.h>
#endif

#include <string.h>

namespace librvngabw
{
Table::Table(TableManager &manager, const librevenge::RVNGPropertyList &xPropList, int xId) :
	m_manager(manager), m_propList(xPropList), m_xId(xId), m_content(), m_rowHeights(),
	m_actualRow(0), m_actualColumn(0),
	m_isRowOpened(false), m_isCellOpened(false),
	m_cellInsertedList()
{
}

Table::~Table()
{
}

void Table::resizeCellInserted(int col, int row)
{
	if (row < 0 || col<0)
	{
		RVNGABW_DEBUG_MSG(("Table::resizeCellInserted: oops bad row or col\n"));
		return;
	}
	if (row < (int) m_cellInsertedList.size() &&
	        col < (int) m_cellInsertedList[size_t(row)].size())
		return;
	if (row >= (int) m_cellInsertedList.size())
		m_cellInsertedList.resize(size_t(row+1));
	for (size_t r=0; r<m_cellInsertedList.size(); ++r)
	{
		if (col >= (int) m_cellInsertedList[r].size())
			m_cellInsertedList[r].resize(size_t(col+1), false);
	}
}

bool Table::openRow(const librevenge::RVNGPropertyList &propList)
{
	if (m_isRowOpened)
	{
		RVNGABW_DEBUG_MSG(("Table::openRow: a row is already open\n"));
		return false;
	}
	m_isRowOpened=true;
	double val;
	if (propList["style:row-height"] && getInchValue(*propList["style:row-height"], val))
		m_rowHeights.push_back(val);
	else if (propList["style:min-row-height"] && getInchValue(*propList["style:min-row-height"], val))
		m_rowHeights.push_back(val);
	else
		m_rowHeights.push_back(-0);
	return true;
}

bool Table::closeRow()
{
	if (!m_isRowOpened)
	{
		RVNGABW_DEBUG_MSG(("Table::closeRow: no row is open\n"));
		return false;
	}
	m_isRowOpened=false;
	++m_actualRow;
	return true;
}

void Table::checkTable(int &xId)
{
	librevenge::RVNGPropertyList cellList, cellPropList;
	// first set all border to none
	for (int i=0; i<4; ++i)
	{
		static char const *(what[])= {"left", "right", "top", "bot"};
		librevenge::RVNGString tmp;
		tmp.sprintf("%s-style", what[i]);
		cellPropList.insert(tmp.cstr(), 0);
	}
	for (size_t r=0; r<m_cellInsertedList.size(); ++r)
	{
		for (size_t c=0; c<m_cellInsertedList[r].size(); ++c)
		{
			if (m_cellInsertedList[r][c]) continue;
			cellList.insert("xid", ++xId);

			cellPropList.insert("left-attach", (int) c);
			cellPropList.insert("right-attach", (int) c+1);
			cellPropList.insert("top-attach", (int) r);
			cellPropList.insert("bottom-attach", (int) r+1);
			librevenge::RVNGPropertyListVector cellPropVector;
			cellPropVector.append(cellPropList);
			cellList.insert("props", cellPropVector);

			m_content.push_back(new TagOpenElement("cell", cellList));
			m_content.push_back(new TagCloseElement("cell"));
		}
	}
}

bool Table::openCell(const librevenge::RVNGPropertyList &propList, int &xId, bool isSpreadsheetCell)
{
	if (!m_isRowOpened || m_isCellOpened)
	{
		RVNGABW_DEBUG_MSG(("Table::openCell: can not open a cell\n"));
		return false;
	}
	int spanCol=1, spanRow=1;
	if (propList["librevenge:column"])
		m_actualColumn=propList["librevenge:column"]->getInt();
	if (propList["librevenge:row"])
		m_actualRow=propList["librevenge:row"]->getInt();
	if (propList["table:number-columns-spanned"] && propList["table:number-columns-spanned"]->getInt()>0)
		spanCol=propList["table:number-columns-spanned"]->getInt();
	if (propList["table:number-rows-spanned"] && propList["table:number-rows-spanned"]->getInt()>0)
		spanRow=propList["table:number-rows-spanned"]->getInt();
	librevenge::RVNGPropertyList cellList, cellPropList;
	cellList.insert("xid", ++xId);
	if (m_actualColumn>=0)
	{
		cellPropList.insert("left-attach", m_actualColumn);
		cellPropList.insert("right-attach", m_actualColumn+spanCol);
	}
	if (m_actualRow>=0)
	{
		cellPropList.insert("top-attach", m_actualRow);
		cellPropList.insert("bottom-attach", m_actualRow+spanRow);
	}

	// other cell's data
	if (propList["fo:background-color"])
		cellPropList.insert("background-color", propList["fo:background-color"]->getStr());
	appendBorderProperties(propList,cellPropList,false);

	if (!cellPropList.empty())
	{
		librevenge::RVNGPropertyListVector cellPropVector;
		cellPropVector.append(cellPropList);
		cellList.insert("props", cellPropVector);
	}
	m_content.push_back(new TagOpenElement("cell", cellList));
	m_isCellOpened=true;

	resizeCellInserted(m_actualColumn+spanCol-1, m_actualRow+spanRow-1);
	for (int r=m_actualRow; r<m_actualRow+spanRow; ++r)
	{
		if (r<0) continue;
		for (int c=m_actualColumn; c<m_actualColumn+spanCol; ++c)
		{
			if (c<0) continue;
			m_cellInsertedList[size_t(r)][size_t(c)]=true;
		}
	}

	if (!isSpreadsheetCell) return true;

	//
	// look if the propList contains a value, if so add it in the cell
	//
	if (!propList["librevenge:value-type"]) return true;

	// now we need to retrieve the value in proplist
	std::string valueType(propList["librevenge:value-type"]->getStr().cstr());
	if (valueType=="double" || valueType=="scientific") valueType="float";
	else if (valueType=="percent") valueType="percentage";

	std::stringstream s;
	s.imbue(std::locale("C")); // be sure that we use standart double
	if (propList["librevenge:value"] && (valueType=="float" || valueType=="percentage" || valueType=="currency"))
	{
		if (valueType=="percentage")
			s << 100. * propList["librevenge:value"]->getDouble() << "%";
		else
		{
			// if (valueType=="currency") m_impl->m_stream << "$";
			s << propList["librevenge:value"]->getDouble();
		}
	}
	else if (propList["librevenge:value"] && (valueType=="bool" || valueType=="boolean"))
	{
		if (propList["librevenge:value"]->getDouble()<0||propList["librevenge:value"]->getDouble()>0)
			s<< "true";
		else
			s<< "false";
	}
	else if (valueType=="date" || valueType=="time")
	{
		// checkme
		struct tm time;
		time.tm_sec=time.tm_min=time.tm_hour=0;
		time.tm_mday=time.tm_mon=1;
		time.tm_year=100;
		time.tm_wday=time.tm_yday=time.tm_isdst=-1;
		char buf[256];
		if (valueType=="date")
		{
			time.tm_mday=propList["librevenge:day"] ? propList["librevenge:day"]->getInt() : 1;
			time.tm_mon=propList["librevenge:month"] ? propList["librevenge:month"]->getInt()-1 : 0;
			time.tm_year=propList["librevenge:year"] ? propList["librevenge:year"]->getInt()-1900 : 100;
			// changeme: use numbering format
			if (mktime(&time)!=-1 && strftime(buf, 256, "%m/%d/%y", &time))
				s << buf;
		}
		else
		{
			time.tm_hour=propList["librevenge:hours"] ? propList["librevenge:hours"]->getInt() : 0;
			time.tm_min=propList["librevenge:minutes"] ? propList["librevenge:minutes"]->getInt() : 0;
			time.tm_sec=propList["librevenge:seconds"] ? propList["librevenge:seconds"]->getInt() : 0;
			// changeme: use numbering format
			if (mktime(&time)!=-1 && strftime(buf, 256, "%H:%M:%S", &time))
				s << buf;
		}
	}
	else if (valueType!="string" && valueType!="text")
	{
		RVNGABW_DEBUG_MSG(("Table::openCell: unexpected value type: %s\n", valueType.c_str()));
		return true;
	}
	if (s.str().empty())
		return true;

	librevenge::RVNGPropertyList paraList;
	if (propList["fo:text-align"])
		paraList.insert("fo:text-align", propList["fo:text-align"]->getStr());
	librevenge::RVNGString sName= m_manager.getSpanManager().findOrAdd(paraList);
	TagOpenElement *pParaOpenElement = new TagOpenElement("p");
	pParaOpenElement->addAttribute("style", sName.cstr());
	pParaOpenElement->addAttribute("xid", ++xId);
	m_content.push_back(pParaOpenElement);

	sName= m_manager.getSpanManager().findOrAdd(propList);
	TagOpenElement *pSpanOpenElement = new TagOpenElement("c");
	pSpanOpenElement->addAttribute("style", sName.cstr());
	pSpanOpenElement->addAttribute("xid", ++xId);
	m_content.push_back(pSpanOpenElement);
	m_content.push_back(new TextElement(s.str().c_str()));
	m_content.push_back(new TagCloseElement("c"));

	m_content.push_back(new TagCloseElement("p"));
	return true;
}

bool Table::closeCell()
{
	if (!m_isCellOpened)
	{
		RVNGABW_DEBUG_MSG(("Table::closeCell: no cell are opened\n"));
		return false;
	}
	m_isCellOpened=false;
	++m_actualColumn;
	m_content.push_back(new TagCloseElement("cell"));
	return true;
}

void Table::write(DocumentElementVector &storage) const
{
	librevenge::RVNGPropertyList propList, propPropList;
	propList.insert("xid", m_xId);

	if (m_propList["table:align"] && m_propList["table:align"]->getStr()=="margins" && m_propList["fo:margin-left"])
		propPropList.insert("table-column-leftpos", m_propList["fo:margin-left"]->getStr());

	const librevenge::RVNGPropertyListVector *columns = m_propList.child("librevenge:table-columns");
	if (columns && columns->count())
	{
		librevenge::RVNGString cols;
		for (unsigned long i=0; i<columns->count(); ++i)
		{
			if ((*columns)[i]["style:column-width"])
				cols.append((*columns)[i]["style:column-width"]->getStr());
			cols.append("/");
		}
		propPropList.insert("table-column-props", cols);
	}
	bool hasRowHeights=false;
	for (size_t i=0; i<m_rowHeights.size(); ++i)
	{
		if (m_rowHeights[i]<=0) continue;
		hasRowHeights=true;
		break;
	}
	if (hasRowHeights)
	{
		librevenge::RVNGString rows;
		for (size_t i=0; i<m_rowHeights.size(); ++i)
		{
			if (m_rowHeights[i]>0)
			{
				std::stringstream s;
				s.imbue(std::locale("C")); // be sure that we use standart double
				s << m_rowHeights[i] << "in";
				rows.append(s.str().c_str());
			}
			rows.append("/");
		}
		propPropList.insert("table-row-heights", rows);
	}
	if (!propPropList.empty())
	{
		librevenge::RVNGPropertyListVector propPropVector;
		propPropVector.append(propPropList);
		propList.insert("props", propPropVector);
	}
	storage.push_back(new TagOpenElement("table", propList));
	// will empty mpContent, which is ok
	const_cast<DocumentElementVector &>(m_content).appendTo(storage);
	storage.push_back(new TagCloseElement("table"));
}

bool Table::insertCoveredCell(const librevenge::RVNGPropertyList &propList, int &)
{
	if (!m_isRowOpened || m_isCellOpened)
	{
		RVNGABW_DEBUG_MSG(("Table::insertCoveredCell: can not open a cell\n"));
		return false;
	}
	if (propList["librevenge:column"])
		m_actualColumn=propList["librevenge:column"]->getInt();
	if (propList["librevenge:row"])
		m_actualRow=propList["librevenge:row"]->getInt();
	return true;
}

TableManager::TableManager(SpanStyleManager &spanManager, ParagraphStyleManager &paragraphManager) :
	m_spanManager(spanManager), m_paragraphManager(paragraphManager), m_tableOpenedList()
{
}

TableManager::~TableManager()
{
}

void TableManager::clean()
{
	m_tableOpenedList.clear();
}

Table *TableManager::openTable(const librevenge::RVNGPropertyList &xPropList, int &xId)
{
	shared_ptr<Table> table(new Table(*this, xPropList, ++xId));
	m_tableOpenedList.push_back(table);
	return table.get();
}

bool TableManager::closeTable()
{
	if (m_tableOpenedList.empty())
	{
		RVNGABW_DEBUG_MSG(("TableManager::oops: no table are opened\n"));
		return false;
	}
	m_tableOpenedList.pop_back();
	return true;
}

}
/* vim:set shiftwidth=4 softtabstop=4 noexpandtab: */
