/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/* librvngabw
 * Version: MPL 2.0 / LGPLv2.1+
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * Major Contributor(s):
 * Copyright (C) 2002-2003 William Lachance (wrlach@gmail.com)
 * Copyright (C) 2004 Fridrich Strba (fridrich.strba@bluewin.ch)
 *
 * For minor contributions see the git repository.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU Lesser General Public License Version 2.1 or later
 * (LGPLv2.1+), in which case the provisions of the LGPLv2.1+ are
 * applicable instead of those above.
 *
 * For further information visit http://libwpd.sourceforge.net
 */

/* "This product is not manufactured, approved, or supported by
 * Corel Corporation or Corel Corporation Limited."
 */

#ifndef _TABLE_HXX_
#define _TABLE_HXX_

#include <librvngabw/librvngabw.hxx>

// for shared_ptr
#include "FilterInternal.hxx"
#include "DocumentElement.hxx"

namespace librvngabw
{
class TableManager;

/** basic class used to store a table */
class Table
{
public:
	//! constructor
	Table(TableManager &manager, const librevenge::RVNGPropertyList &xPropList, int xId);
	//! destructor
	virtual ~Table();
	//! return the table storage
	DocumentElementVector *getStorage()
	{
		return &m_content;
	}
	//! add table content to storage
	void write(DocumentElementVector &storage) const;
	//! check that the table is complete, if not insert empty cell
	void checkTable(int &xId);

	//! open a row
	bool openRow(const librevenge::RVNGPropertyList &propList);
	//! close a row
	bool closeRow();
	//! returns true if a row is opened
	bool isRowOpened() const
	{
		return m_isRowOpened;
	}
	//! open a cell
	bool openCell(const librevenge::RVNGPropertyList &propList, int &xId, bool isSpreadsheetCell=false);
	//! close a cell
	bool closeCell();
	//! insert a covered cell
	bool insertCoveredCell(const librevenge::RVNGPropertyList &propList, int &xId);
	//! returns true if a cell is opened
	bool isCellOpened() const
	{
		return m_isCellOpened;
	}
protected:
	//! check that the cell list contains row and column
	void resizeCellInserted(int col, int row);
private:
	//! the table manager
	TableManager &m_manager;
	//! the table proplist
	librevenge::RVNGPropertyList m_propList;
	//! the table id
	int m_xId;
	//! the table content
	DocumentElementVector m_content;
	//! the list of row heights
	std::vector<double> m_rowHeights;
	//! the actual row
	int m_actualRow;
	//! the actual column
	int m_actualColumn;
	//! flag to know if a row is opened
	bool m_isRowOpened;
	//! flag to know if a cell is opened
	bool m_isCellOpened;
	//! list of set cell
	std::vector< std::vector<bool> > m_cellInsertedList;
	// disable copying
	Table(const Table &);
	Table &operator=(const Table &);
};

class ParagraphStyleManager;
class SpanStyleManager;

/** class which stores all the tables
 */
class TableManager
{
public:
	//! constructor
	TableManager(SpanStyleManager &spanManager, ParagraphStyleManager &paragraphManager);
	//! destructor
	virtual ~TableManager();
	//! clean all data
	void clean();
	//! returns true if a table is opened
	bool isTableOpened() const
	{
		return !m_tableOpenedList.empty();
	}
	//! returns the actual table
	Table *getActualTable()
	{
		if (m_tableOpenedList.empty()) return 0;
		return m_tableOpenedList.back().get();
	}
	//! returns the actual table
	Table const *getActualTable() const
	{
		if (m_tableOpenedList.empty()) return 0;
		return m_tableOpenedList.back().get();
	}
	//! open a table and update the list of elements
	Table *openTable(const librevenge::RVNGPropertyList &xPropList, int &xId);
	//! close a table
	bool closeTable();

	//! returns the span manager
	SpanStyleManager &getSpanManager()
	{
		return m_spanManager;
	}

	//! returns the paragraph manager
	ParagraphStyleManager &getParagraphManager()
	{
		return m_paragraphManager;
	}

private:
	//! the span manager
	SpanStyleManager &m_spanManager;
	//! the paragraph manager
	ParagraphStyleManager &m_paragraphManager;
	//! the list of opened table
	std::vector<shared_ptr<Table> > m_tableOpenedList;

	// disable copying
	TableManager(const TableManager &);
	TableManager &operator=(const TableManager &);
};
}

#endif

/* vim:set shiftwidth=4 softtabstop=4 noexpandtab: */
